/* $Id: module.c,v 1.9 2004/02/20 12:56:20 armin Exp $
 *
 * divactrl aix configure/unconfigure kernel drivers
 *
 * (c) 2003 Cytronics & Melware
 *
 */

#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/errno.h>
#include <sys/sysmacros.h>
#include <sys/sysconfig.h>
#include <sys/device.h>
#include <sys/mode.h>

#include <sys/cfgodm.h>
#include <sys/cfgdb.h>
#include <cf.h>

#define KEXTENSIONDIR "/usr/lib/drivers"
#define DIVACTRLBIN "/sbin/divactrl"

#define ADAPTER_NUMS	12

#define TASK_NONE 0
#define TASK_LOAD 1
#define TASK_UNLOAD 2
#define TASK_QUERY 3
#define TASK_INFO 4 

static char		*modulename;
static int		task = TASK_NONE;
static int		verbose = 1;
static struct cfg_load	cload;
static struct cfg_dd	cfgdd;
static int		majorno;
static int		minorno;
static dev_t		devno;
static unsigned char	load_file[256];
static unsigned char	*devpath, *devname;
static unsigned int	devmode;

/*
 * show usage
 */
static void usage(void)
{
	if (!verbose)
		return;
	
	fprintf(stderr, "usage: divactrl aixmodule -v [-l|-u|-q|-i] -f <module-name>\n");
	fprintf(stderr, "       -v                  : be non-verbose\n");
	fprintf(stderr, "       -l -f <module-name> : load module\n");
	fprintf(stderr, "       -u -f <module-name> : unload module\n");
	fprintf(stderr, "       -q -f <module-name> : query module\n");
	fprintf(stderr, "       -i -f <module-name> : show info\n");
}

/*
 * get information of module
 */
static int do_info(void)
{
	if (cload.kmid == 0) {
		if (verbose)
			fprintf(stderr, "Error, module %s is not loaded\n", modulename);
		return(1);
	}
	/* TODO */
	return (0);
}

/*
 * remove devpath
 */
static void remove_devpath(char *dp)
{
	int i;
	char t[80];
	
	unlink(dp);
	
	if (!strcmp(dp, "/dev/Divas")) {
		unlink("/dev/DivasINFO");
		for (i = 1; i < ADAPTER_NUMS; i++) {
			sprintf(t, "/dev/DivasAdapter%d", i);
			unlink(t);
		}
	}
}

/*
 * load and config
 */
static int do_load(void)
{
	int i;
	char t[80];
	int rc;
	dev_t mdevno;
	
	if (cload.kmid != 0) {
		if (verbose)
			fprintf(stderr, "Error, module %s already loaded\n", modulename);
		return(1);
	}

	if ((majorno = genmajor(devname)) == -1) {
		if (verbose)
			fprintf(stderr, "Error, cannot genmajor\n");
		return(E_MAJORNO);
	}
	minorno = 0;
	
	devno = makedev(majorno, minorno);

	remove_devpath(devpath);
	if ((rc = mknod(devpath, _S_IFCHR | devmode, devno)) == -1) {
		reldevno(devname, TRUE);
		if (verbose)
	                fprintf(stderr, "MKNOD error\n");
		return(rc);
	}
	chmod(devpath, devmode);

	if (!strcmp(devpath, "/dev/Divas")) {
		mdevno = makedev(majorno, 255);
		mknod("/dev/DivasINFO", _S_IFCHR | devmode, mdevno);
		chmod("/dev/DivasINFO", 0644);
		for (i = 1; i < ADAPTER_NUMS; i++) {
			sprintf(t, "/dev/DivasAdapter%d", i);
			mdevno = makedev(majorno, i);
			mknod(t, _S_IFCHR | devmode, mdevno);
		}
	}
	
	if (sysconfig(SYS_SINGLELOAD, &cload, sizeof(struct cfg_load))) {
		reldevno(devname, TRUE);
		remove_devpath(devpath);
		if (verbose)
	                fprintf(stderr, "LOAD error\n");
		return(1);
	}
	if (verbose)
		printf("Module %s loaded--kmid is %d\n", modulename, cload.kmid);
	
	/* configure the module */
	cfgdd.kmid   = cload.kmid;
	cfgdd.cmd    = CFG_INIT;
	cfgdd.devno  = devno;
	if (sysconfig(SYS_CFGDD, &cfgdd, sizeof(cfgdd))) {
		reldevno(devname, TRUE);
		remove_devpath(devpath);
		sysconfig(SYS_KULOAD, &cload, sizeof(struct cfg_load));
		if (verbose)
			fprintf(stderr, "INIT error\n");
		return(1);
	}
	return(0);
}

/*
 * term and unload
 */
static int do_unload(void)
{
	if (cload.kmid == 0) {
		if (verbose)
			fprintf(stderr, "Error, module %s is not loaded\n", modulename);
		return(1);
	}

	/* check for QVPD */
	cfgdd.kmid   = cload.kmid;
	cfgdd.cmd    = CFG_QVPD;
	cfgdd.devno  = devno;
	if (sysconfig(SYS_CFGDD, &cfgdd, sizeof(cfgdd))) {
		if (verbose)
			fprintf(stderr, "Module reports <in use>\n");
		return(1);
	}
	
	/* unconfigure the module */
	cfgdd.kmid   = cload.kmid;
	cfgdd.cmd    = CFG_TERM;
	cfgdd.devno  = devno;
	if (sysconfig(SYS_CFGDD, &cfgdd, sizeof(cfgdd))) {
		if (verbose)
			fprintf(stderr, "TERM error\n");
	}
	reldevno(devname, TRUE);
	remove_devpath(devpath);
	if (sysconfig(SYS_KULOAD, &cload, sizeof(struct cfg_load))) {
		if (verbose)
			fprintf(stderr, "UNLOAD error\n");
		return(1);
	}
	if (verbose)
		printf("Module %s unloaded\n", modulename);
	return(0);
}

/*
 * Query
 */
static int do_query(void)
{
	int ret = 0;
	
	if (cload.kmid == 0) {
		if (verbose)
			printf("Module %s is not loaded\n", modulename);
		ret = 1;
	} else {
		if (verbose)
			printf("Module %s is loaded with kmid=%d\n", modulename, cload.kmid);
	}
	return(ret);
}


/*
 * (un)load all modules needed for system
 */
static int do_system(void)
{
	int ret = 0;
	int fd = 0;
	int lines = 0, c = 0, i;
	char page[1024];

	switch (task) {
		case TASK_UNLOAD:
			if ((fd = open("/dev/DivasINFO", O_RDONLY)) >= 0) {
				c = read(fd, page, 1024);
				for (i = 0; i < c; i++) {
					if (page[i] == '\n') lines++;
				}
				close(fd);
			}
			if (lines <= 1) {
				system(DIVACTRLBIN" aixmodule -v -u -f idi >/dev/null 2>&1");
				system(DIVACTRLBIN" aixmodule -v -u -f divas >/dev/null 2>&1");
				system(DIVACTRLBIN" aixmodule -v -u -f maint >/dev/null 2>&1");
				system(DIVACTRLBIN" aixmodule -v -u -f capi >/dev/null 2>&1");
				system(DIVACTRLBIN" aixmodule -v -u -f kernelcapi >/dev/null 2>&1");
				system(DIVACTRLBIN" aixmodule -v -u -f didd >/dev/null 2>&1");
			}
			break;

		case TASK_LOAD:
			system(DIVACTRLBIN" aixmodule -v -l -f didd >/dev/null 2>&1");
			system(DIVACTRLBIN" aixmodule -v -l -f kernelcapi >/dev/null 2>&1");
			system(DIVACTRLBIN" aixmodule -v -l -f capi >/dev/null 2>&1");
			system(DIVACTRLBIN" aixmodule -v -l -f divas >/dev/null 2>&1");
			system(DIVACTRLBIN" aixmodule -v -l -f idi >/dev/null 2>&1");
			break;

		default:
			/* unknown task for this command */
			ret = 1;
			break;
	}
	return(ret);
}

/*
 * main
 */
int aixmodule_main (int argc, char** argv)
{
	int c;
	struct stat   statbuf;
	int ret = 0;

	/* parse options */
	while ((c = getopt(argc, argv, "vluqihf:")) != EOF) {
		switch(c) {
			case 'v':
				verbose = 0;
				break;
			case 'l':
				task = TASK_LOAD;
				break;
			case 'u':
				task = TASK_UNLOAD;
				break;
			case 'q':
				task = TASK_QUERY;
				break;
			case 'i':
				task = TASK_INFO;
				break;
			case 'f':
				modulename = strdup(optarg);
				break;
			case 'h':
			case '?':
				usage();
				return(0);
		}
	}
	if (argc > optind) {
		if (verbose)
			fprintf(stderr, "extra parameter, command line syntax error\n");
		return(-1);
	}
	if (!modulename) {
		if (verbose)
			fprintf(stderr, "module-name missing\n");
		return(-1);
	}
	if (task == TASK_NONE) {
		if (verbose)
			fprintf(stderr, "no task specified\n");
		return(-1);
	}

	if (getuid() != 0) {
		if (verbose)
			fprintf(stderr, "root-permission is required\n");
		return(EACCES);
	}
	
	/* create full module name */
	if (!strcasecmp(modulename, "divas")) {
		sprintf(load_file, "%s/%s", KEXTENSIONDIR, "divas");
		devpath = "/dev/Divas";
		devname = "Divas";
		devmode = 0600;
	} else
	if (!strcasecmp(modulename, "didd")) {
		sprintf(load_file, "%s/%s", KEXTENSIONDIR, "divadidd");
		devpath = "/dev/DivaDIDD";
		devname = "DivaDIDD";
		devmode = 0600;
	} else
	if (!strcasecmp(modulename, "capi")) {
		sprintf(load_file, "%s/%s", KEXTENSIONDIR, "divacapi");
		devpath = "/dev/DivaCAPI";
		devname = "DivaCAPI";
		devmode = 0600;
	} else
	if (!strcasecmp(modulename, "idi")) {
		sprintf(load_file, "%s/%s", KEXTENSIONDIR, "diva_idi");
		devpath = "/dev/DivasIDI";
		devname = "DivasIDI";
		devmode = 0666;
	} else
	if (!strcasecmp(modulename, "maint")) {
		sprintf(load_file, "%s/%s", KEXTENSIONDIR, "diva_mnt");
		devpath = "/dev/DivasMAINT";
		devname = "DivasMAINT";
		devmode = 0600;
	} else
	if (!strcasecmp(modulename, "kernelcapi")) {
		sprintf(load_file, "%s/%s", KEXTENSIONDIR, "kernelcapi");
		devpath = "/dev/capi20";
		devname = "capi20";
		devmode = 0666;
	} else
	if (!strcmp(modulename, "SYSTEM")) {
		return(do_system());
	} else
	{
		if (verbose)
			fprintf(stderr, "unknown module name '%s'\n", modulename);
		return(EINVAL);
	}

	/* check if file exists */
	if (stat(load_file, &statbuf) != 0) {
		if (verbose)
			fprintf(stderr, "%s not found\n", load_file);
		return(errno);
	}

	cload.path = load_file;
	cload.libpath = NULL;
	cload.kmid = 0;

	/* check if module is loaded */
	if (sysconfig(SYS_QUERYLOAD, &cload, sizeof(struct cfg_load))) {
		if (verbose)
			fprintf(stderr, "QUERY error\n");
		return(-1);
	}

	if (odm_initialize() == -1) {
		/* initialization failed */
		exit(E_ODMINIT);
	}

	switch (task) {
		case TASK_QUERY:
			ret = do_query();
			break;

		case TASK_UNLOAD:
			ret = do_unload();
			break;

		case TASK_LOAD:
			ret = do_load();
			break;

		case TASK_INFO:
			ret = do_info();
			break;

		default:
			if (verbose)
				fprintf(stderr, "Unknown task, error\n");
			ret = 1;
	}
	odm_terminate();

	return(ret);
}

